#include "test/Service.h"
#include "test/Tester.h"

#include <boost/format.hpp>
#include <boost/range/adaptors.hpp>
#include <nlohmann/json.hpp>

#include <chrono>
#include <functional>
#include <thread>


namespace xs { namespace test {

namespace detail {
    std::string toString(const nlohmann::json& object)
    {
        return object.is_string() ? object.get<std::string>()
                                  : object.dump();
    }
}

Tester::Tester(const boost::filesystem::path& file)
: config_(file)
, server_(
    "xs-test", { }, { }, 0, nullptr
)
{
    registerServices();
}

void Tester::registerServices() noexcept
{
    acceptCommands_.clear();
    acceptEvents_.clear();

    using namespace std::placeholders;
    for (const auto& service : config_.getServices()) {
        if (service.isCommand()) {
            server_.onCmd(
                service.name_,
                std::bind(&Tester::onCommand, this, _1, _2)
            );
            acceptCommands_[service.name_] = &service;
        } else {
            server_.onEvent(
                service.name_,
                std::bind(&Tester::onEvent, this, _1, _2)
            );
            acceptEvents_[service.name_] = &service;
        }
    }
} 

std::string Tester::onCommand(const std::string& name, const std::string& params) noexcept
{
    const auto found = acceptCommands_.find(name);
    const auto* service = std::get<1>(*found);
    const auto& response = service->response_;

    /// TODO: keep track the status
    return detail::toString(response);
}

void Tester::onEvent(const std::string& name, const std::string& params) noexcept
{
    const auto found = acceptEvents_.find(name);
    const auto* service = std::get<1>(*found);
    const auto& response = service->response_;

    /// TODO: keep track the status
}

void Tester::run() noexcept
{
    for (const auto& e : config_.getExpectations()) {
        const auto& command = e.name_;
        const auto& expect  = e.value_;

        const auto response = server_.runCmd(command, "");
        bool asExpect = false;
        if (expect.is_string()) {
            asExpect = (response == expect.get<std::string>());
        } else {
            try {
                asExpect = (nlohmann::json::parse(response) == expect);
            } catch (...) {

            }
        }

        if (asExpect) {
            std::cout << "[PASS] run command: " << command << std::endl;
        } else {
            std::cout << "[FAILED] run command: " << command << " with unexpected response: "
                      << "\"" << response << "\"(real) vs " << detail::toString(expect) << "(expect)" << std::endl;
        }
    }
}

} }
